(* Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast

module StringMap = Map.Make(String)

let translate (globals, functions) =
  let context = L.global_context () in
  let llctx = L.global_context () in
  let lib_buffer = L.MemoryBuffer.of_file "lib/utils.bc" in
  let llm = Llvm_bitreader.parse_bitcode llctx lib_buffer in
  let the_module = L.create_module context "Sipl"
  and i32_t  = L.i32_type  context
  and i8_t   = L.i8_type   context
  and i1_t   = L.i1_type   context
  and void_t = L.void_type context
  and string_t = L.pointer_type (L.i8_type context)
  and float_t = L.double_type context
  and image_t = L.pointer_type (match L.type_by_name llm "struct.img" with
    None -> raise (Failure "struct.img doesn't defined.")
  | Some x -> x)
  and matrix_t = L.pointer_type (match L.type_by_name llm "struct.matrix" with
    None -> raise (Failure "struct.matrix doesn't defined.")
  | Some x -> x)
  and pixel_t = L.pointer_type (match L.type_by_name llm "struct.pixel" with
    None -> raise (Failure "struct.pixel doesn't defined.")
  | Some x -> x)
in


  let ltype_of_typ = function
      A.Int -> i32_t
    | A.Bool -> i1_t
    | A.Image -> image_t
    | A.Matrix -> matrix_t
    | A.Pixel -> pixel_t
    | A.Void -> void_t 
    | A.Float -> float_t
    | A.String -> string_t in

  (* Declare each global variable; remember its value in a map *)
  let global_vars =
    let global_var m (t, n) =
      let init = L.const_int (ltype_of_typ t) 0
      in StringMap.add n (L.define_global n init the_module, t) m in
    List.fold_left global_var StringMap.empty globals in

  (* Declare printf(), which the print built-in function will call *)
  let printf_t = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func = L.declare_function "printf" printf_t the_module in

  (* Declare the built-in printbig() function *)
  let printbig_t = L.function_type i32_t [| i32_t |] in
  let printbig_func = L.declare_function "printbig" printbig_t the_module in

  (* Declare built-in functions*)
  let printImage_t = L.function_type i32_t [| image_t |] in
  let printImage_func = L.declare_function "printImage" printImage_t the_module in

  let printMatrix_t = L.function_type i32_t [| matrix_t |] in
  let printMatrix_f = L.declare_function "printMatrix" printMatrix_t the_module in

  let printPixel_t = L.function_type i32_t [| pixel_t |] in
  let printPixel_f = L.declare_function "printPixel" printPixel_t the_module in

   (* Maths -------------------------------------------------------------*)

  let int2float_t = L.function_type float_t [| i32_t |] in
  let int2float_f = L.declare_function "int2float" int2float_t the_module in

  let float2int_t = L.function_type i32_t [| float_t |] in
  let float2int_f = L.declare_function "float2int" float2int_t the_module in

  let power_t = L.function_type float_t [| float_t; float_t |] in
  let power_f = L.declare_function "power" power_t the_module in

  let matAccess_t = L.function_type float_t [| matrix_t; i32_t; i32_t |] in
  let matAccess_f = L.declare_function "matAccess" matAccess_t the_module in

  let matAssign_t = L.function_type float_t [| matrix_t; i32_t; i32_t; float_t |] in
  let matAssign_f = L.declare_function "matAssign" matAssign_t the_module in

  let imgAccess_t = L.function_type i32_t [| image_t; i32_t; i32_t; string_t;|] in
  let imgAccess_f = L.declare_function "imgAccess" imgAccess_t the_module in

  let imgAssign_t = L.function_type i32_t [| image_t; i32_t; i32_t; string_t; i32_t|] in
  let imgAssign_f = L.declare_function "imgAssign" imgAssign_t the_module in

  let pixAccess_t = L.function_type pixel_t [| image_t; i32_t; i32_t|] in
  let pixAccess_f = L.declare_function "pixAccess" pixAccess_t the_module in

  let pixAssign_t = L.function_type i32_t [| image_t; i32_t; i32_t; pixel_t|] in
  let pixAssign_f = L.declare_function "pixAssign" pixAssign_t the_module in

  let initMatrix_t      = L.function_type matrix_t [| L.pointer_type i8_t; i32_t; i32_t |] in
  let initMatrix_func   = L.declare_function "initMatrix" initMatrix_t the_module in

  let imgWidth_t = L.function_type i32_t [| image_t |] in
  let imgWidth_f = L.declare_function "imgWidth" imgWidth_t the_module in

  let matWidth_t = L.function_type i32_t [| matrix_t |] in
  let matWidth_f = L.declare_function "matWidth" matWidth_t the_module in

  let imgHeight_t = L.function_type i32_t [| image_t |] in
  let imgHeight_f = L.declare_function "imgHeight" imgHeight_t the_module in

  let matHeight_t = L.function_type i32_t [| matrix_t |] in
  let matHeight_f = L.declare_function "matHeight" matHeight_t the_module in

  let kernel_t = L.function_type matrix_t [| string_t |] in
  let kernel_f = L.declare_function "kernel" kernel_t the_module in

  let imgread_t = L.function_type image_t [| string_t |] in
  let imgread_func = L.declare_function "imgread" imgread_t the_module in
    
  let imgwrite_t = L.function_type i32_t [| image_t; i32_t; string_t |] in
  let imgwrite_func = L.declare_function "imgwrite" imgwrite_t the_module in

  let convImg_t = L.function_type image_t [| image_t; matrix_t |] in
  let convImg_f = L.declare_function "convImg" convImg_t the_module in

  let changeGrey_t = L.function_type i32_t [| image_t |] in
  let changeGrey_f = L.declare_function "changeGrey" changeGrey_t the_module in

  let rotateImage_t = L.function_type i32_t [| image_t |] in
  let rotateImage_f = L.declare_function "rotateImage" rotateImage_t the_module in

  let flipImage_t = L.function_type i32_t [| image_t; i32_t |] in
  let flipImage_f = L.declare_function "flipImage" flipImage_t the_module in



  let elementAdd_t = L.function_type image_t [| image_t; i32_t |] in
  let elementAdd_f = L.declare_function "elementAdd" elementAdd_t the_module in

  let elementSub_t = L.function_type image_t [| image_t; i32_t |] in
  let elementSub_f = L.declare_function "elementSub" elementSub_t the_module in

  let elementMult_t = L.function_type image_t [| image_t; float_t |] in
  let elementMult_f = L.declare_function "elementMult" elementMult_t the_module in

  let imgAdd_t = L.function_type image_t [| image_t; image_t |] in
  let imgAdd_f = L.declare_function "imgAdd" imgAdd_t the_module in

  let imgCopy_t = L.function_type image_t [| image_t |] in
  let imgCopy_f = L.declare_function "imgCopy" imgCopy_t the_module in

  let imgCreate_t = L.function_type image_t [| i32_t; i32_t |] in
  let imgCreate_f = L.declare_function "imgCreate" imgCreate_t the_module in

  let matCreate_t = L.function_type matrix_t [| i32_t; i32_t |] in
  let matCreate_f = L.declare_function "matCreate" matCreate_t the_module in

(*             Pixel Functions *)

  let pixAddPix_t = L.function_type pixel_t [| pixel_t; pixel_t |] in
  let pixAddPix_f = L.declare_function "pixAddPix" pixAddPix_t the_module in

  let pixAddInt_t = L.function_type pixel_t [| pixel_t; i32_t |] in
  let pixAddInt_f = L.declare_function "pixAddInt" pixAddInt_t the_module in

  let pixSubInt_t = L.function_type pixel_t [| pixel_t; i32_t |] in
  let pixSubInt_f = L.declare_function "pixSubInt" pixSubInt_t the_module in

(*   let pixMultInt_t = L.function_type pixel_t [| pixel_t; i32_t |] in
  let pixMultInt_f = L.declare_function "pixMultInt" pixMultInt_t the_module in *)

  let pixMultFloat_t = L.function_type pixel_t [| pixel_t; float_t |] in
  let pixMultFloat_f = L.declare_function "pixMultFloat" pixMultFloat_t the_module in

  (* Define each function (arguments and return type) so we can call it *)
  let function_decls =
    let function_decl m fdecl =
      let name = fdecl.A.fname
      and formal_types =
	Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.A.formals)
      in let ftype = L.function_type (ltype_of_typ fdecl.A.typ) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m in
    List.fold_left function_decl StringMap.empty functions in
  
  (* Fill in the body of the given function *)
  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.A.fname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in

    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder in
    let string_format_str = L.build_global_stringptr "%s\n" "fmt" builder in
    let float_format_str = L.build_global_stringptr "%.3f\n" "fmt" builder in

    (* Construct the function's "locals": formal arguments and locally
       declared variables.  Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals" map *)
    let local_vars =
      let add_formal m (t, n) p = L.set_value_name n p;
	let local = L.build_alloca (ltype_of_typ t) n builder in
	ignore (L.build_store p local builder);
	StringMap.add n (local, t) m in

      let add_local m (t, n) =
	let local_var = L.build_alloca (ltype_of_typ t) n builder
	in StringMap.add n (local_var, t) m in

      let formals = List.fold_left2 add_formal StringMap.empty fdecl.A.formals
          (Array.to_list (L.params the_function)) in
      List.fold_left add_local formals fdecl.A.locals in

    (* Return the value for a variable or formal argument *)
    let lookup n = try StringMap.find n local_vars
                   with Not_found -> StringMap.find n global_vars
    in

    let idx n m          = [| L.const_int i32_t n; L.const_int i32_t m |] in

    let get_mptr m b = 
        let arr_ptr = L.build_in_bounds_gep m (idx 0 0) "build_in_bounds_gep" b in 
         L.build_bitcast arr_ptr (L.pointer_type i8_t) "mat_ptr" b 
    in

    let build_matrix l expr builder = 
        let rows = List.length l in
        let cols = List.length (List.hd l) in
        let flat_list = List.fold_left List.append [] l in
        let list_of_const = List.map fst (List.map (expr builder) flat_list) in
        let flat_array = Array.of_list list_of_const in
        let mat = L.const_array float_t flat_array in
        let var = L.build_alloca (L.array_type float_t (cols * rows)) "mat" builder in 
        ignore (L.build_store mat var builder);
        L.build_call initMatrix_func [| (get_mptr var builder); (L.const_int i32_t rows); 
                                (L.const_int i32_t cols) |] "initMatrix" builder
    in

    let get_string = function 
      A.StringLit s -> s
      | _ -> raise (Failure ("Get String of a none string"))

    in

    (* Construct code for an expression; return its value *)
    let rec expr builder = function
	A.Literal i -> (L.const_int i32_t i, A.Int)
      | A.StringLit s -> (L.build_global_stringptr s s builder, A.String)
      | A.BoolLit b -> (L.const_int i1_t (if b then 1 else 0), A.Bool)
      | A.FloatLit f -> (L.const_float float_t f, A.Float)
      | A.MatrixLit l -> (build_matrix l expr builder, A.Matrix)
      | A.Noexpr -> (L.const_int i32_t 0, A.Void)
      | A.Id s -> (L.build_load (fst (lookup s)) s builder, snd (lookup s)) 
      | A.Binop (e1, op, e2) ->
        	  let (e1', typ1) = expr builder e1
        	  and (e2', typ2) = expr builder e2 in
        	  (match op with
        	    A.Add     -> 
                (match typ1 with
                    A.Int     -> 
                        (match typ2 with
                            A.Int     -> (L.build_add e1' e2' "tmp" builder, typ1)
                          | A.Float   -> (L.build_fadd (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Float)
                          | A.Image   -> (L.build_call elementAdd_f [| e2' ; e1' |] "elementAdd" builder, typ2)
                          | A.Pixel   -> (L.build_call pixAddInt_f [| e2' ; e1' |] "pixAddInt" builder, typ2)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | A.Float   -> 
                        (match typ2 with
                            A.Float     -> (L.build_fadd e1' e2' "tmp" builder, typ1)
                          | A.Int   -> (L.build_fadd e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Float)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | A.Image   -> 
                        (match typ2 with
                            A.Int     -> (L.build_call elementAdd_f [| e1' ; e2' |] "elementAdd" builder, typ1)
                          | A.Image   -> (L.build_call imgAdd_f [| e1' ; e2' |] "elementAdd" builder, typ1)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | A.Pixel   -> 
                        (match typ2 with
                            A.Int     -> (L.build_call pixAddInt_f [| e1' ; e2' |] "pixAddInt" builder, typ1)
                          | A.Pixel   -> (L.build_call pixAddPix_f [| e1' ; e2' |] "pixAddPix" builder, typ1)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | _         -> raise (Failure (" unsupported operation "))
                )
        	  | A.Sub     -> 
                (match typ1 with
                    A.Int     -> 
                        (match typ2 with
                            A.Int     -> (L.build_sub e1' e2' "tmp" builder, typ1)
                          | A.Float   -> (L.build_fsub (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Float)
                          | _         -> raise (Failure (" unsupported operation "))
                        )

                  | A.Float   -> 
                        (match typ2 with
                            A.Float     -> (L.build_fsub e1' e2' "tmp" builder, typ1)
                          | A.Int   -> (L.build_fsub e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Float)
                          | _         -> raise (Failure (" unsupported operation "))
                        )

                  | A.Image   -> 
                        (match typ2 with
                            A.Int     -> (L.build_call elementSub_f [| e1' ; e2' |] "elementSub" builder, typ1)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | A.Pixel   -> 
                        (match typ2 with
                            A.Int     -> (L.build_call pixSubInt_f [| e1' ; e2' |] "pixSubInt" builder, typ1)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | _         -> raise (Failure (" unsupported operation "))
                )
        	  | A.Mult    -> 
                (match typ1 with
                    A.Int     -> 
                        (match typ2 with
                            A.Int     -> (L.build_mul e1' e2' "tmp" builder, A.Int)
                          | A.Float   -> (L.build_fmul (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Float)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | A.Float   -> 
                      (match typ2 with
                            A.Float   -> (L.build_fmul e1' e2' "tmp" builder, typ1)
                          | A.Int     -> (L.build_fmul e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Float)
                          | A.Image   -> (L.build_call elementMult_f [| e2' ; e1' |] "elementMult" builder, typ2)
                          | A.Pixel   -> (L.build_call pixMultFloat_f [| e2' ; e1' |] "pixMultFloat" builder, typ2)
                          | _         -> raise (Failure (" unsupported operation "))
                      )
                  | A.Image   -> 
                      (match typ2 with
                            A.Float     -> (L.build_call elementMult_f [| e1' ; e2' |] "elementMult" builder, typ1)
                          | _         -> raise (Failure (" unsupported operation "))
                      )
                  | A.Pixel   -> 
                      (match typ2 with
                            A.Float     -> (L.build_call pixMultFloat_f [| e1' ; e2' |] "pixMultFloat" builder, typ1)
                          | _         -> raise (Failure (" unsupported operation "))
                      )
                  | _         -> raise (Failure (" unsupported operation "))
                )
            | A.Div     -> 
                (match typ1 with
                    A.Int     -> 
                      (match typ2 with
                            A.Int     -> (L.build_sdiv e1' e2' "tmp" builder, A.Int)
                          | A.Float   -> (L.build_fdiv (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Float)
                          | _         -> raise (Failure (" unsupported operation "))
                        )

                  | A.Float   -> 
                      (match typ2 with
                            A.Float     -> (L.build_fdiv e1' e2' "tmp" builder, A.Float)
                          | A.Int   -> (L.build_fdiv e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Float)
                          | _         -> raise (Failure (" unsupported operation "))
                        )
                  | _         -> raise (Failure (" unsupported operation "))
                )
        	  | A.And     -> (L.build_and e1' e2' "tmp" builder, A.Bool)
        	  | A.Or      -> (L.build_or e1' e2' "tmp" builder, A.Bool)
        	  | A.Equal   -> 
                (match typ1 with
                    A.Int     -> (L.build_icmp L.Icmp.Eq e1' e2' "tmp" builder, A.Bool)
                  | A.Float   -> (L.build_fcmp L.Fcmp.Ueq e1' e2' "tmp" builder, A.Bool)
                  | _         -> raise (Failure (" unsupported operation "))
                )

        	  | A.Neq     -> 
                (match typ1 with
                    A.Int     -> (L.build_icmp L.Icmp.Ne e1' e2' "tmp" builder, A.Bool)
                  | A.Float   -> (L.build_fcmp L.Fcmp.Une e1' e2' "tmp" builder, A.Bool)
                  | _         -> raise (Failure (" unsupported operation "))
                )
                
        	  | A.Less    -> 
                (match typ1 with
                    A.Int     -> 
                      (match typ2 with
                            A.Int -> (L.build_icmp L.Icmp.Slt e1' e2' "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Ult (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | A.Float   -> 
                      (match typ2 with
                            A.Int -> (L.build_fcmp L.Fcmp.Ult e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Ult e1' e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )

                  
                  | _         -> raise (Failure (" unsupported operation "))
                )
        	  | A.Leq     -> 
                (match typ1 with
                    A.Int     -> 
                      (match typ2 with
                            A.Int -> (L.build_icmp L.Icmp.Sle e1' e2' "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Ule (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | A.Float   -> 
                      (match typ2 with
                            A.Int -> (L.build_fcmp L.Fcmp.Ule e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Ule e1' e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | _         -> raise (Failure (" unsupported operation "))
                )
        	  | A.Greater -> 
                (match typ1 with
                    A.Int     -> 
                      (match typ2 with
                            A.Int -> (L.build_icmp L.Icmp.Sgt e1' e2' "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Ugt (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | A.Float   -> 
                      (match typ2 with
                            A.Int -> (L.build_fcmp L.Fcmp.Ugt e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Ugt e1' e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | _         -> raise (Failure (" unsupported operation "))
                )
        	  | A.Geq     -> 
                (match typ1 with
                    A.Int     -> 
                      (match typ2 with
                            A.Int -> (L.build_icmp L.Icmp.Sge e1' e2' "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Uge (L.build_call int2float_f [| e1' |] "int2float" builder) e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | A.Float   -> 
                    (match typ2 with
                            A.Int -> (L.build_fcmp L.Fcmp.Uge e1' (L.build_call int2float_f [| e2' |] "int2float" builder) "tmp" builder, A.Bool)
                            | A.Float -> (L.build_fcmp L.Fcmp.Uge e1' e2' "tmp" builder, A.Bool)
                            | _ -> raise (Failure (" unsupported operation "))
                      )
                  | _         -> raise (Failure (" unsupported operation "))
                )
              | A.Conv  ->
                (match typ1 with 
                    A.Image -> ( match typ2 with
                        A.Matrix   -> (L.build_call convImg_f [|e1' ; e2' |] "convImg" builder, A.Image)
                        | _ ->raise (Failure("unsupport operation"))
                    )

                    | A.Matrix -> ( match typ2 with
                        A.Image   -> (L.build_call convImg_f [|e2' ; e1' |] "convImg" builder, A.Image)
                        | _ -> raise (Failure("unsupport operation"))
                    )
                    | _ -> raise (Failure("unsupport operation"))
                )
        	  ) 
      | A.Unop(op, e) ->
        	  let (e', typ') = expr builder e in
        	  ((match op with
        	         A.Neg     -> L.build_neg
                  | A.Not     -> L.build_not) e' "tmp" builder, typ')
      | A.MatAccess (s, r, c) ->
              let row = fst (expr builder r) in
              let col = fst (expr builder c) in
              let mat = L.build_load (fst (lookup s)) s builder in 
              (L.build_call matAccess_f [| mat; row; col |] "matAccess" builder , A.Float)
      | A.MatAssign (s, r, c, e) ->
              let (e', typ') = expr builder e in
              let row = fst (expr builder r) in
              let col = fst (expr builder c) in
              let mat = L.build_load (fst (lookup s)) s builder in 
              (match typ' with
                A.Float -> (L.build_call matAssign_f [| mat; row; col; e' |] "matAssign" builder , A.Int)
                | A.Int -> 
                (L.build_call matAssign_f [| mat; row; col; (L.build_call int2float_f [| e' |] "int2float" builder) |] "matAssign" builder , A.Int)
                | _ -> raise (Failure("unsupport Matrix Assign"))
              )
              
      | A.ImgAccess (s, r, c, sr) ->
              let row = fst (expr builder r) in
              let col = fst (expr builder c) in
              let str = fst (expr builder sr) in
              let img = L.build_load (fst (lookup s)) s builder in 
              let flag = get_string sr in
              if (flag = "a") then (L.build_call pixAccess_f [| img; row; col |] "pixAccess" builder , A.Pixel)
              else (L.build_call imgAccess_f [| img; row; col; str |] "imgAccess" builder , A.Int)
              

      | A.ImgAssign (s, r, c, sr, e) ->
              let e' = fst (expr builder e) in
              let row = fst (expr builder r) in
              let col = fst (expr builder c) in
              let str = fst (expr builder sr) in
              let img = L.build_load (fst (lookup s)) s builder in 
              let flag = get_string sr in
              if (flag = "a") then (L.build_call pixAssign_f [| img; row; col; e' |] "pixAssign" builder , A.Int)
              else (L.build_call imgAssign_f [| img; row; col; str; e' |] "imgAssign" builder , A.Int)
      | A.GetWidth (id) ->
            let (s, typ) =  lookup id in
            let id' = L.build_load s id builder in
              (match typ with
                A.Image -> (L.build_call imgWidth_f [| id'|] "img" builder, A.Int)
                | A.Matrix -> (L.build_call matWidth_f [| id'|] "mat" builder, A.Int)
                | _ -> raise (Failure("unsupport operation"))
                )
                 
      | A.GetHeight (id) ->
            let (s, typ) =  lookup id in
            let id' = L.build_load s id builder in
              (match typ with
                A.Image -> (L.build_call imgHeight_f [| id'|] "img" builder, A.Int)
                | A.Matrix -> (L.build_call matHeight_f [| id'|] "mat" builder, A.Int)
                | _ -> raise (Failure("unsupport operation"))
                )
      | A.Assign (s, e) -> let (e', typ') = expr builder e and (l, ltyp) = lookup s in 
              (match (ltyp, typ') with
                (A.Float, A.Int) -> 
                    ignore (L.build_store (L.build_call int2float_f [| e' |] "int2float" builder) l builder); (e', ltyp)
                | _ -> ignore (L.build_store e' l builder ); (e', typ')
              )
	                   
      | A.Call ("print", [e]) | A.Call ("printb", [e]) -> 
	            (L.build_call printf_func [| int_format_str ; fst (expr builder e) |] "printf" builder , A.Int)
      | A.Call ("prints", [e]) ->
              (L.build_call printf_func [| string_format_str ; fst (expr builder e) |] "printf" builder, A.Int)
      | A.Call ("printfloat", [e]) ->
              (L.build_call printf_func [| float_format_str ; fst (expr builder e) |] "printf" builder, A.Int)
      | A.Call ("printbig", [e]) ->
	             (L.build_call printbig_func [| fst (expr builder e) |] "printbig" builder, A.Int)
      | A.Call ("printImage", [e]) ->
              (L.build_call printImage_func [| fst (expr builder e) |] "printImage" builder, A.Int)
      | A.Call ("printMatrix", [e]) ->
              (L.build_call printMatrix_f [| fst (expr builder e) |] "printMatrix" builder, A.Int)
      | A.Call ("printPixel", [e]) ->
              (L.build_call printPixel_f [| fst (expr builder e) |] "printPixel" builder, A.Int)
      | A.Call ("int2float", [e]) -> 
              (L.build_call int2float_f [| fst (expr builder e) |] "int2float" builder , A.Float)
      | A.Call ("float2int", [e]) -> 
              (L.build_call float2int_f [| fst (expr builder e) |] "float2int" builder , A.Int)
      | A.Call ("power", [e1; e2]) -> 
              (L.build_call power_f [| fst (expr builder e1); fst (expr builder e2) |] "power" builder , A.Float)
      | A.Call ("kernel", [e]) ->
              (L.build_call kernel_f [| fst (expr builder e) |] "kernel" builder, A.Matrix)
      | A.Call ("matCreate", [e1; e2]) ->
              (L.build_call matCreate_f [| fst (expr builder e1); fst (expr builder e2) |] "matCreate" builder, A.Matrix)
      | A.Call ("changeGrey", [img]) ->
              (L.build_call changeGrey_f [| fst (expr builder img) |] "changeGrey" builder, A.Int)
      | A.Call ("rotateImage", [img]) ->
              (L.build_call rotateImage_f [| fst (expr builder img) |] "rotateImage" builder, A.Int)
      | A.Call ("flipImage", [img; flag]) ->
              (L.build_call flipImage_f [| fst (expr builder img); fst (expr builder flag)|] "flipImage" builder, A.Int)
      | A.Call ("imgCopy", [img]) ->
              (L.build_call imgCopy_f [| fst (expr builder img) |] "imgCopy" builder, A.Image)
      | A.Call ("imgCreate", [e1; e2]) ->
              (L.build_call imgCreate_f [| fst (expr builder e1); fst (expr builder e2) |] "imgCreate" builder, A.Image)
      | A.Call ("imgread", [filename]) ->
              (L.build_call imgread_func [| fst (expr builder filename)|] "imgread" builder, A.Image)
      | A.Call ("imgwrite", [img; flag; filename]) ->
              (L.build_call imgwrite_func [| fst (expr builder img); fst (expr builder flag); fst (expr builder filename) |] "imgwrite" builder, A.Int)
      | A.Call (f, act) ->
         let (fdef, fdecl) = StringMap.find f function_decls in
	 let actuals = List.rev (List.map fst (List.map (expr builder) (List.rev act))) in
	 let result = (match fdecl.A.typ with A.Void -> ""
                                            | _ -> f ^ "_result") in
         (L.build_call fdef (Array.of_list actuals) result builder, fdecl.A.typ)
    in

    
    
    (* Invoke "f builder" if the current block doesn't already
       have a terminal (e.g., a branch). *)
    let add_terminal builder f =
      match L.block_terminator (L.insertion_block builder) with
	Some _ -> ()
      | None -> ignore (f builder) in
	
    (* Build the code for the given statement; return the builder for
       the statement's successor *)
    let rec stmt builder = function
	A.Block sl -> List.fold_left stmt builder sl
      | A.Expr e -> ignore (expr builder e); builder
      | A.Return e -> ignore (match fdecl.A.typ with
	           A.Void -> L.build_ret_void builder
	             | _ -> L.build_ret (fst (expr builder e)) builder); builder
      | A.If (predicate, then_stmt, else_stmt) ->
         let bool_val = fst (expr builder predicate) in
	 let merge_bb = L.append_block context "merge" the_function in

	 let then_bb = L.append_block context "then" the_function in
	 add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
	   (L.build_br merge_bb);

	 let else_bb = L.append_block context "else" the_function in
	 add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
	   (L.build_br merge_bb);

	 ignore (L.build_cond_br bool_val then_bb else_bb builder);
	 L.builder_at_end context merge_bb

      | A.While (predicate, body) ->
	  let pred_bb = L.append_block context "while" the_function in
	  ignore (L.build_br pred_bb builder);

	  let body_bb = L.append_block context "while_body" the_function in
	  add_terminal (stmt (L.builder_at_end context body_bb) body)
	    (L.build_br pred_bb);

	  let pred_builder = L.builder_at_end context pred_bb in
	  let bool_val = fst (expr pred_builder predicate) in

	  let merge_bb = L.append_block context "merge" the_function in
	  ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
	  L.builder_at_end context merge_bb

      | A.For (e1, e2, e3, body) -> stmt builder
	    ( A.Block [A.Expr e1 ; A.While (e2, A.Block [body ; A.Expr e3]) ] )
    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (A.Block fdecl.A.body) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.A.typ with
        A.Void -> L.build_ret_void
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body functions;
  the_module
